"""
The methods to run the parts of the form compiler

Should also contain the entrypoint methods.
"""
from __future__ import absolute_import

def read_ufl(uflfile):
    from ufl.algorithms.formfiles import read_ufl_file, interpret_ufl_namespace
    from ufl.algorithms import compute_form_data

    # Read the given ufl file and execute it
    uflcode = read_ufl_file(uflfile)
    namespace = {}
    try:
        exec "from dune.perftool.ufl.execution import *\n" + uflcode in namespace
    except:
        import os
        basename = os.path.splitext(os.path.basename(uflfile))[0]
        basename = "{}_debug".format(basename)
        pyname = "{}.py".format(basename)
        print pyname
        pycode = "#!/usr/bin/env python\nfrom dune.perftool.ufl.execution import *\nset_level(DEBUG)\n" + uflcode
        with file(pyname, "w") as f:
            f.write(pycode)
        raise SyntaxError("Not a valid ufl file, dumped a debug script: {}".format(pyname))

    # Extract the form from the given data
    data = interpret_ufl_namespace(namespace)
    form = data.forms[0]
    form = compute_form_data(form).preprocessed_form

    # We do not expect more than one form
    assert len(data.forms) == 1

    # TODO As validity is implemented right now, this is nonsense.
    # Might as well formulate it as a try statement!
#     from dune.perftool.ufl.validity import check_validity
#     if not check_validity(form):
        # apply some transformations unconditionally!
    from dune.perftool.ufl.transformations import transform_form
    from dune.perftool.ufl.transformations.splitarguments import split_arguments
    from dune.perftool.ufl.transformations.indexpushdown import pushdown_indexed
    from dune.perftool.ufl.transformations.reindexing import reindexing

    form = transform_form(form, pushdown_indexed)
    form = transform_form(form, reindexing)
    form = transform_form(form, split_arguments)

    return form


def generate_driver(form, filename):
    # The driver module uses a global variable for ease of use
    from dune.perftool.pdelab.driver import set_form
    set_form(form)

    # The vtkoutput is the generating method that triggers all others.
    # Alternatively, one could use the `solve` method.
    from dune.perftool.pdelab.driver import vtkoutput
    vtkoutput()

    from dune.perftool.generation import retrieve_cache_items

    # Get all preambles for the driver and sort them.
    driver_content = [i[1] for i in sorted(retrieve_cache_items("preamble"), key=lambda x : x[0])]

    # And flatten out those, that contained nested lists
    def flatjoin(l):
        if isinstance(l, str):
            return "\n" + l
        return "".join(flatjoin(i) for i in l)

    from cgen import LiteralBlock
    driver = LiteralBlock(flatjoin(driver_content))

    # Write the file.
    f = open(filename, 'w')
    f.write("\n".join(retrieve_cache_items("include")))
    f.write("\n\nvoid driver(int argc, char** argv)\n")
    f.write("\n".join(driver.generate()))

    # Reset the caching data structure
    from dune.perftool.generation import delete_cache
    delete_cache()


def compile_form():
    from dune.perftool.options import get_option
    form = read_ufl(get_option("uflfile"))

    if get_option("driver_file"):
        generate_driver(form, get_option("driver_file"))

    if get_option("operator_file"):
        from dune.perftool.pdelab.localoperator import generate_localoperator
        generate_localoperator(form, get_option("operator_file"))
