from __future__ import absolute_import
from ufl.algorithms import MultiFunction


class EliminateArguments(MultiFunction):
    """ Application scope: Take a valid accumulation term and remove the modified argument from it! """
    call = MultiFunction.__call__

    def __call__(self, o):
        from dune.perftool.ufl.modified_terminals import ModifiedArgumentExtractor
        from dune.perftool.ufl.rank import ufl_rank

        self.arguments = ModifiedArgumentExtractor()(o)
        e = self.call(o)

        # Catch the case that the entire expression vanished!
        if e is None:
            from ufl.classes import IntValue
            e = IntValue(1)

        return (e, self.arguments)

    def expr(self, o):
        if o in self.arguments:
            return None
        else:
            # Evaluate the multi function applied to the operands
            newop = tuple(self.call(op) for op in o.ufl_operands)
            # Find out whether an operand vanished. If so, the class needs special treatment.
            if None in newop:
                raise NotImplementedError("Operand vanished: {} needs special treatment in EliminateArguments".format(type(o)))
            return self.reuse_if_untouched(o, *newop)

    def sum(self, o):
        assert len(o.ufl_operands) == 2

        op0 = self.call(o.ufl_operands[0])
        op1 = self.call(o.ufl_operands[1])

        if op0 and op1:
            return self.reuse_if_untouched(o, op0, op1)
        # One term vanished, so there is no sum anymore
        else:
            if op0 or op1:
                # Return the term that did not vanish
                return op0 if op0 else op1
            else:
                # This entire sum vanished!!!
                return None

    # The handler for product is equal to the sum handler
    product = sum

    def index_sum(self, o):
        op = self.call(o.ufl_operands[0])
        if op:
            return self.reuse_if_untouched(o, op, o.ufl_operands[1])
        else:
            return None
