from __future__ import absolute_import
from ufl.algorithms import MultiFunction


def check_validity(uflexpr):
    """ check an UFL expression (usually an integrand) for
    compatibility with the dune-pdelab code generation tool chain.

    The assumptions made are the following:
    - The expression is the sum of subtrees, where the number of test
      function terms in such subtree is equal the rank of the entire expression.
    """
    from ufl import Form
    from ufl.classes import Expr

    def check(term):
        try:
            from dune.perftool.ufl.transformations.extract_accumulation_terms import split_into_accumulation_terms
            split_into_accumulation_terms(term)
            return True
        except AssertionError:
            from dune.perftool.ufl.transformations import print_expression
            print_expression(term)
            return False

    if isinstance(uflexpr, Form):
        for integral in uflexpr.integrals():
            if not check(integral.integrand()):
                return False
        return True

    if isinstance(uflexpr, Expr):
        return check(uflexpr)

    raise TypeError("Unknown object type in check_validity: {}".format(type(uflexpr)))
